import io
import re
import time
from threading import Thread
from tests.__init__ import *
from psplpy.other_utils import *


def test_output_capture():
    capture = OutputCapture()
    print("This is stdout1")
    print("This is stdout2")
    print("This is stderr", file=sys.stderr)
    # test get_output() and stop_capture()
    stdout, stderr = capture.get_output_and_stop()
    assert stdout == "This is stdout1\nThis is stdout2\n", stdout
    assert stderr == "This is stderr\n", stderr
    # test after clear(), get two empty strings
    capture.clear()
    stdout, stderr = capture.get_output()
    assert stdout == '', stdout
    assert stderr == '', stderr
    # test cannot get anything after stop_capture() and will output on console
    print("This is stdout")
    print("This is stderr", file=sys.stderr)
    assert stdout == '', stdout
    assert stderr == '', stderr
    # test can capture after restart_capture()
    capture.restart_capture()
    print("This is stdout")
    print("This is stderr", file=sys.stderr)
    stdout, stderr = capture.get_output_and_stop()
    assert stdout == "This is stdout\n", stdout
    assert stderr == "This is stderr\n", stderr


def test_input_mock():
    input_mock = InputMock()
    # test normal input
    input_mock.input('hello')
    assert input() == 'hello'
    input_mock.input('hello\n\n\n')
    assert input() == 'hello'
    # test the EOFError will occur when input an empty string
    try:
        input_mock.input('')
        input()
    except EOFError as e:
        print(str(e))
    else:
        assert False
    # test the EOFError will not occur when set auto_line_wrapping=True
    input_mock = InputMock(auto_line_wrapping=True)
    try:
        input_mock.input('')
    finally:
        assert input() == ''
    # restore the stdin
    input_mock.restore()


def test_class_property():
    class Config:
        PROJECT_DIR = Path(__file__).parent.parent

        @ClassProperty
        def SRC_DIR(cls):
            return cls.PROJECT_DIR / 'src'

        @ClassProperty
        def DATA_DIR(cls):
            return cls.PROJECT_DIR / 'data'

    # test whether the 'ClassProperty' can return the correct value
    assert Config.SRC_DIR == Path('/home/a/src/psplpyProject/src')
    # test assigning a new variable, thus hiding the 'ClassProperty' instance
    Config.SRC_DIR = Path(__file__).parent
    assert Config.SRC_DIR == Path('/home/a/src/psplpyProject/tests')
    # test assigning a new 'ClassProperty' instance
    Config.SRC_DIR = ClassProperty(lambda cls: cls.PROJECT_DIR / '1')
    assert Config.SRC_DIR == Path('/home/a/src/psplpyProject/1')


def test_is_sys():
    assert is_sys(is_sys.LINUX) is True
    assert is_sys(is_sys.WINDOWS) is False


def test_recursive_convert():
    data = [1, (2, [3, 4])]
    assert recursive_convert(data, to=list) == [1, [2, [3, 4]]]
    assert recursive_convert(data, to=tuple) == (1, (2, (3, 4)))


def test_split_list():
    lst = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    assert split_list(lst, 4) == [[1, 2, 3], [4, 5], [6, 7], [8, 9]]


def test_get_key_from_value():
    dct = {'a': 1, 'b': 2, 'c': 1, 'd': 1}
    assert get_key_from_value(dct, 1) == 'a'
    assert get_key_from_value(dct, 2, find_all=True) == ['b']
    assert get_key_from_value(dct, 1, find_all=True) == ['a', 'c', 'd']
    assert get_key_from_value(dct, 3, allow_null=True) is None
    assert get_key_from_value(dct, 3, find_all=True, allow_null=True) is None
    try:
        get_key_from_value(dct, 3)
    except KeyError as e:
        print(e)
    else:
        assert False


def test_get_env():
    assert get_env('SERVICE') == 'psplpy'
    try:
        get_env('S')
    except KeyError as e:
        print(str(e))
    else:
        raise AssertionError


def test_perf_counter():
    p = PerfCounter()
    time.sleep(0.1)
    assert int(p.elapsed()) == int(0.1 * 1000)
    time.sleep(0.2)
    p.show('perf')


def test_timer():
    # basic test
    interval = 0.1
    timer = Timer(interval=interval)
    start_time = time.time()
    for i in range(10):
        timer.wait()
        elapsed = time.time() - start_time
        assert elapsed < (i + 2) * interval, elapsed
        print(elapsed)
    # test warning and error
    output_capture = OutputCapture()
    timer = Timer(interval=interval, warning_multiple=1, error_multiple=2)
    try:
        for i in range(1000):
            timer.wait()
            time.sleep(interval * 1.75)
    except TimeoutError as e:
        print(str(e))
        stdout, stderr = output_capture.get_output_and_stop()
        assert re.match(r'Current delay: 0\.2.s\n', stdout), repr(stdout)
        assert re.match(r'Warning, current delay: 0\.1.s\n'
                        r'Warning, current delay: 0\.2.s\n', stderr), repr(stderr)
    else:
        assert False


def test_timeout_checker():
    def checker(timeout: float, unique_id: Any = None, check_time: float = None,
                func: str = 'ret_false', end_loop_time: float = 10000):
        t_start = time.time()
        while getattr(TimeoutChecker(timeout, unique_id=unique_id), func)():
            if time.time() - t_start > end_loop_time:
                break
            time.sleep(0.01)
        print(time.time() - t_start)
        assert int(time.time() - t_start) == check_time or timeout

    Thread(target=checker, kwargs={'timeout': 1, 'unique_id': 1}).start()
    timeout = 2
    Thread(target=checker, kwargs={'timeout': 2, 'unique_id': 2}).start()
    time.sleep(timeout + 0.5)

    kwargs = {'timeout': 2, 'unique_id': 1, 'check_time': 1, 'func': 'raise_err', 'end_loop_time': 1}
    Thread(target=checker, kwargs=kwargs).start()
    try:
        checker(timeout=1, unique_id=2, func='raise_err')
    except Exception as e:
        assert isinstance(e, TimeoutError)


def test_screenshot_maker():
    pass
    # ScreenShotMaker()


def test_gen_passwd():
    print(gen_passwd())


def tests():
    print('test other utils')
    test_output_capture()
    test_input_mock()
    test_class_property()
    test_is_sys()
    test_recursive_convert()
    test_split_list()
    test_get_key_from_value()
    test_get_env()
    test_perf_counter()
    test_timer()
    test_timeout_checker()
    test_screenshot_maker()
    test_gen_passwd()


if __name__ == '__main__':
    tests()



from tests.__init__ import *
from psplpy.dynamic_compose import DynamicCompose

dc_rc_dir = rc_dir / 'dynamic_compose'
dc_compose_dir = dc_rc_dir / '.compose'
compose_dumped_file = dc_compose_dir / 'compose-dumped.yml'
dockerfile_dumped_file = dc_compose_dir / 'dumped-Dockerfile'
dockerfile_dumped_file2 = dc_compose_dir / 'dumped-Dockerfile2'


def lines_rstrip(text: str) -> str:
    lines = text.splitlines()
    stripped_lines = [line.rstrip() for line in lines]
    result = '\n'.join(stripped_lines)
    return result


def tests():
    dc = DynamicCompose(cwd=dc_rc_dir)
    dc.env['GPU'] = '''    
    deploy:
      resources:
        reservations:
          devices:
            - driver: nvidia
              count: all
              capabilities: [gpu]'''
    dc.env['IMAGE'] = 'python:3.10'
    dc.env['software'] = 'git'
    print(dc.env)

    dc.fd_all().up()

    assert compose_dumped_file.read_text().strip() == lines_rstrip((dc.cwd / 'compose.yml').read_text().strip())
    assert dockerfile_dumped_file.read_text().strip() == lines_rstrip((dc.cwd / 'Dockerfile').read_text().strip())
    assert dockerfile_dumped_file2.read_text().strip() == lines_rstrip((dc.cwd / 'Dockerfile2').read_text().strip())


if __name__ == '__main__':
    tests()



import os
import re
import django
from django.conf import settings
from django.db import models, connections
from psplpy.image_det import TxtProc
from pathlib import Path


class DbBackend:
    SQLITE3 = 'django.db.backends.sqlite3'
    POSTGRESQL = 'django.db.backends.postgresql'
    MYSQL = 'django.db.backends.mysql'


class Database:
    def __init__(self, engine: str = DbBackend.SQLITE3, name=None, user=None, password=None,
                 host='localhost', port=None, app_label: str = '', **django_settings):
        databases = {
            'default': {
                'ENGINE': engine,
                'NAME': name,
                'USER': user,
                'PASSWORD': password,
                'HOST': host,
                'PORT': port,
                'APP_LABEL': app_label,
            }
        }

        # Update the settings with the custom DATABASES dictionary
        settings_dict = {'DATABASES': databases}
        settings_dict.update(django_settings)
        settings.configure(**settings_dict)
        # Initialize Django
        django.setup()

        # Create the custom base model
        class CustomBaseModel(models.Model):
            class Meta:
                app_label = databases['default']['APP_LABEL']
                abstract = True

            @classmethod
            def init(cls, new_name=None) -> None:
                Database.change_table_name(cls, new_name)
                Database.create_table(cls)
                Database.update_table(cls)

        self.Model = CustomBaseModel

    @staticmethod
    def change_table_name(model, new_name=None) -> None:
        if new_name is None:
            new_name = TxtProc.camel_to_snake(model.__name__)
        model._meta.db_table = new_name

    @staticmethod
    # Create a table if it doesn't exist
    def create_table(model) -> None:
        with connections['default'].schema_editor() as schema_editor:
            if model._meta.db_table not in connections['default'].introspection.table_names():
                schema_editor.create_model(model)

    @staticmethod
    # Update table if you added fields (doesn't drop fields as far as i know, which i was too afraid to implement)
    def update_table(model) -> None:
        with connections['default'].schema_editor() as schema_editor:
            if model._meta.db_table not in connections['default'].introspection.table_names():
                raise ValueError(f'Table "{model._meta.db_table}" not found.')
            else:
                # Get the database columns
                database_columns = connections['default'].introspection.get_table_description(
                    connections['default'].cursor(), model._meta.db_table)
                database_column_names = [column.name for column in database_columns]
                # Check if each field in the model exists in the database table
                for field in model._meta.fields:
                    if field.column not in database_column_names:
                        # Add the new column to the table
                        schema_editor.add_field(model, field)


class DjangoInit:
    def __init__(self, project_dir: str | Path = None, database: str = None):
        self.project_dir = Path(project_dir) if project_dir else os.getcwd()
        self.database = database
        self.manage_py_path = self._find_manage_py()
        self.project_name = self._extract_project_name()
        self.settings_py_path = self.project_dir / self.project_name / 'settings.py'
        self.settings_str = self.settings_py_path.read_text(encoding='utf-8')

    def _find_manage_py(self) -> Path:
        manage_py_path = self.project_dir / 'manage.py'
        if manage_py_path.exists():
            return manage_py_path
        raise FileNotFoundError(f'manage.py not found at {manage_py_path}')

    def _extract_project_name(self) -> str:
        content = self.manage_py_path.read_text(encoding='utf-8')
        match = re.search(r"os\.environ\.setdefault\('DJANGO_SETTINGS_MODULE', '([^']+)'\)", content)
        if match:
            return match.group(1).split('.')[0]
        raise ValueError("Cannot extract project name from manage.py")

    def _mkdir(self, dir_path: Path) -> None:
        if dir_path.exists():
            dir_path.mkdir(exist_ok=True)
            print(f'Make dir: {dir_path}')

    def _replace_settings(self, old: str, new: str) -> None:
        if old in self.settings_str:
            self.settings_str = self.settings_str.replace(old, new)
            print(f'Change: {old}')
            print(f'To: {new}')

    def _replace_file(self, file_path: Path, old: str, new: str) -> None:
        if not file_path.exists():
            raise ValueError(f'File does not exist: {file_path}')
        content = file_path.read_text(encoding='utf-8')
        if old in content:
            file_path.write_text(content.replace(old, new), encoding='utf-8')
            print(f'Change: {old}')
            print(f'To: {new}')

    def _set_asgi(self):
        new = f"""'django.contrib.staticfiles',
            'channels',
        ]

        ASGI_APPLICATION = '{self.project_name}.asgi.application'

        CHANNEL_LAYERS = {{
            'default': {{
                'BACKEND': 'channels_redis.core.RedisChannelLayer',
                'CONFIG': {{
                    "hosts": [(get_env('REDIS_HOST'), get_env('REDIS_PORT'))],
                }},
            }},
        }}"""
        self._replace_settings("'django.contrib.staticfiles',\n]", new)
        old = f"""\nfrom django.core.asgi import get_asgi_application

os.environ.setdefault('DJANGO_SETTINGS_MODULE', '{self.project_name}.settings')

application = get_asgi_application()"""
        new = f"""from django.core.asgi import get_asgi_application
from channels.routing import ProtocolTypeRouter, URLRouter
from channels.auth import AuthMiddlewareStack

os.environ.setdefault('DJANGO_SETTINGS_MODULE', '{self.project_name}.settings')

application = ProtocolTypeRouter({{
    "http": get_asgi_application(),
    "websocket": AuthMiddlewareStack(
        URLRouter(
            routing.websocket_urlpatterns
        )
    ),
}})
"""
        self._replace_file(self.project_dir / self.project_name / 'asgi.py', old, new)

    def init(self) -> None:
        # make global templates dir
        self._mkdir(self.project_dir / 'templates')
        self._replace_settings("'DIRS': [],", "'DIRS': [BASE_DIR / 'templates'],")
        # make global static dir
        self._mkdir(self.project_dir / 'static')
        new = "STATIC_URL = 'static/'\nSTATICFILES_DIRS = [BASE_DIR / 'static',]"
        self._replace_settings("STATIC_URL = 'static/'", new)
        # modify allowed hosts
        self._replace_settings("ALLOWED_HOSTS = []", "ALLOWED_HOSTS = ['*']")
        # set_asgi
        self._set_asgi()
        # set database
        if self.database and self.database != DbBackend.SQLITE3:
            db_setting = f"""DATABASES = {{
    'default': {{
        'ENGINE': '{self.database}',
        'NAME': get_env("DB_NAME"),
        'USER': get_env("DB_USER"),
        'PASSWORD': get_env("DB_PASSWORD"),
        'HOST': get_env("DB_HOST"),
        'PORT': get_env("DB_PORT"),
    }}
}}"""
            old = """DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': BASE_DIR / 'db.sqlite3',
    }
}"""
            self._replace_settings(old, db_setting)
            old = ("from pathlib import Path\n\n"
                   "# Build paths inside the project like this: BASE_DIR / 'subdir'.")
            new = ('from pathlib import Path\n'
                   'from psplpy.other_utils import get_env\n'
                   'get_env.env_file = Path(__file__).resolve().parent.parent.parent / ".env"\n\n'
                   "# Build paths inside the project like this: BASE_DIR / 'subdir'.")
            self._replace_settings(old, new)
        # write settings
        self.settings_py_path.write_text(self.settings_str, encoding='utf-8')



import shutil
from datetime import datetime
from zoneinfo import ZoneInfo
from django.utils import timezone
from tests.__init__ import *
from psplpy.django_utils import *


def test_database():
    db = Database(engine=DbBackend.POSTGRESQL, name=get_env('DB_USER'), host=get_env('DB_HOST'), port=get_env('DB_PORT'),
                  user=get_env('DB_USER'), password=get_env('DB_PW'), TIME_ZONE='UTC')
    assert timezone.get_current_timezone() == ZoneInfo('UTC')

    class Test(db.Model):
        name = models.CharField(max_length=200, null=True)
        email = models.EmailField(null=True)
        time = models.DateTimeField(default=timezone.now())

    Test.init()

    naive_time = datetime(2020, 1, 1, 0, 0)
    aware_time = timezone.make_aware(naive_time)
    print(naive_time, aware_time)

    test = Test(name='me', email='test@test.com', time=aware_time)
    test.save()

    all_values = [list(item.values()) for item in Test.objects.all().values()]
    print(all_values)
    assert all_values[0] == [1, 'me', 'test@test.com', aware_time]


def test_django_init():
    rc_path = rc_dir / 'django_utils'
    tmp_path = tmp_dir / 'django_utils'
    if tmp_path.exists():
        shutil.rmtree(tmp_path)
    shutil.copytree(rc_path, tmp_path)
    modified_settings_content = (rc_path / 'test' / 'settings_modified.py').read_text(encoding='utf-8')

    DjangoInit(project_dir=tmp_path, database=DbBackend.POSTGRESQL).init()
    assert (tmp_path / 'test' / 'settings_modified.py').read_text(encoding='utf-8') == modified_settings_content


def tests():
    test_database()
    test_django_init()


if __name__ == '__main__':
    tests()
